#pragma once
#include <exception>

//Copyright (C) cocosimulator.org 2023
//generated by AmsterCHEM's COM CAPE-OPEN Wizard

// an exception class for internal exceptions

class COException : public exception {

	//use a stack buffer (staticText) while we can, avoid anything that throws an exception
	static const int maxStaticText=512;
	wchar_t staticText[maxStaticText];
	wchar_t *dynamicText;
	HRESULT code;

	public:

	COException(const wchar_t *reason,HRESULT _code=ECapeUnknownHR) : code(_code) {
		dynamicText=nullptr;
		SetText(reason,(int)wcslen(reason));
	}

	COException(const wstring &reason,HRESULT _code=ECapeUnknownHR) : code(_code) {
		dynamicText=nullptr;
		SetText(reason.c_str(),(int)reason.size());
	}

	COException(const COException &other) {
		dynamicText=nullptr;
		code=other.code;
		if (other.dynamicText) {
			SetText(other.staticText,(int)wcslen(other.staticText));
		} else {
			#ifdef _MSC_VER
			wcscpy_s(staticText,maxStaticText,other.staticText);
			#else
			wcscpy(staticText,other.staticText);
			#endif
		}
	}

	~COException() {
		if (dynamicText) {
			free(dynamicText);
			dynamicText=nullptr;
		}
	}

	void SetCode(HRESULT _code) {
		code=_code;
	}

	void SetText(const wchar_t *reason,int reasonLen) {
		if (dynamicText) {
			delete []dynamicText;
			dynamicText=nullptr;
		}
		if (reasonLen<maxStaticText) {
			#ifdef _MSC_VER
			wcscpy_s(staticText,maxStaticText,reason);
			#else
			wcscpy(staticText,reason);
			#endif
		} else {
			dynamicText=(wchar_t*)malloc(2*(reasonLen+1));
			if (dynamicText) {
				#ifdef _MSC_VER
				wcscpy_s(dynamicText,reasonLen+1,reason);
				#else
				wcscpy(dynamicText,reason);
				#endif
			} else {
				//copy partly
				memcpy(staticText,reason,2*(maxStaticText-1));
				staticText[maxStaticText-1]=0;
			}
		}
	}

	bool Empty() {
		return ((dynamicText==nullptr)&&(*staticText==0));
	}

	int Length() {
		return (int)wcslen(wwhat());
	}

	void Prepend(const wchar_t *prefix,int prefixLen) {
		if (Empty()) {
			SetText(prefix,prefixLen);
		} else {
			int oldLen=Length();
			if ((prefixLen+oldLen)<maxStaticText) {
				ASSERT(!dynamicText,"unexpected text buffer");
				for (int i=oldLen;i>=0;i--) staticText[i+prefixLen]=staticText[i];
				memcpy(staticText,prefix,2*prefixLen);
			} else {
				const wchar_t *oldText=wwhat();
				wchar_t *newText=(wchar_t *)malloc(2*(prefixLen+oldLen+1));
				if (newText) {
					memcpy(newText,prefix,2*prefixLen);
					memcpy(newText+prefixLen,oldText,2*(oldLen+1));
					if (dynamicText) free(dynamicText);
					dynamicText=newText;
				} else {
					//fit what we can in static buffer
					staticText[maxStaticText-1]=0;
					int n=maxStaticText-1-prefixLen;
					if (n>0) {
						for (int i=n;i>=0;i--) {
							staticText[i+prefixLen]=staticText[i];
						}
					}
					n=prefixLen;
					if (n>(maxStaticText-1)) n=maxStaticText-1;
					memcpy(staticText,prefix,n);
					if (dynamicText) {
						free(dynamicText);
						dynamicText=nullptr;
					}
				}
			}
		}
	}

	void Prepend(const wchar_t *prefix) {
		Prepend(prefix,(int)wcslen(prefix));
	}

	void Prepend(const wstring &prefix) {
		Prepend(prefix.c_str(),(int)prefix.size());
	}

	void Append(const wchar_t *postfix,int postfixLen) {
		if (Empty()) {
			SetText(postfix,postfixLen);
		} else {
			int oldLen=Length();
			if ((postfixLen+oldLen)<maxStaticText) {
				ASSERT(!dynamicText,"unexpected text buffer");
				memcpy(staticText+oldLen,postfix,2*(postfixLen+1));
			} else {
				const wchar_t *oldText=wwhat();
				wchar_t *newText=(wchar_t *)malloc(2*(postfixLen+oldLen+1));
				if (newText) {
					memcpy(newText,oldText,oldLen*2);
					memcpy(newText+oldLen,postfix,2*(postfixLen+1));
					if (dynamicText) free(dynamicText);
					dynamicText=newText;
				} else {
					//fit what we can in static buffer
					if (!dynamicText) { //else do not make shorter
						int n=maxStaticText-1-oldLen;
						memcpy(staticText,postfix,2*n);
						staticText[maxStaticText-1]=0;
					}
				}
			}
		}
	}

	void Append(const wchar_t *postfix) {
		Append(postfix,(int)wcslen(postfix));
	}

	void Append(const wstring &postfix) {
		Append(postfix.c_str(),(int)postfix.size());
	}

	const char *what() const throw() {
		return UnicodeToMultibyte(wwhat());
	}

	const wchar_t *wwhat() const {
		return (dynamicText)?dynamicText:staticText;
	}

	HRESULT Code() const {
		return code;
	}

};

//two macros for C++ exception handling and CAPE-OPEN error handling
//  note that COMaterialException is the exception to throw in case something fails
#define BEGIN_TRY(iface) \
	try {
#define END_TRY(iface) \
	} catch (COException &ex) { \
		SetError(ex.wwhat(),__FUNCTIONW__,iface); \
		return ex.Code(); \
	} catch (exception &ex) { \
		SetError(MultibyteToUnicode(ex.what()),__FUNCTIONW__,iface); \
		return ECapeUnknownHR; \
	} catch (...) { \
		SetError(L"Unhandled exception",__FUNCTIONW__,iface); \
		return ECapeUnknownHR; \
	} \
	return NO_ERROR;
